local abs, acos, sqrt, sin = math.abs, math.acos, math.sqrt, math.sin

-- converted from this website:
-- http://www.euclideanspace.com/maths/algebra/realNormedAlgebra/quaternions/slerp/index.htm
function math3d.Slerp(qa, qb, delta)
	local qa_w, qa_x, qa_y, qa_z = qa:w(), qa:x(), qa:y(), qa:z()
	local qb_w, qb_x, qb_y, qb_z = qb:w(), qb:x(), qb:y(), qb:z()
	
	local cosHalfTheta = qa_w * qb_w + qa_x * qb_x + qa_y * qb_y + qa_z * qb_z
	
	if abs(cosHalfTheta) >= 0.9995 then
		return qa:clone()
	end
	
	if cosHalfTheta < 0 then -- make sure to use the shortest angle
		qb_w, qb_x, qb_y, qb_z = -qb_w, -qb_x, -qb_y, -qb_z
		cosHalfTheta = -cosHalfTheta
	end
	
	local halfTheta = acos(cosHalfTheta)
	local sinHalfTheta = sqrt(1 - cosHalfTheta*cosHalfTheta)
	
	if abs(sinHalfTheta) < 0.001 then
		local w = qa_w * 0.5 + qb_w * 0.5
		local x = qa_x * 0.5 + qb_x * 0.5
		local y = qa_y * 0.5 + qb_y * 0.5
		local z = qa_z * 0.5 + qb_z * 0.5
		
		return quaternion(w, x, y, z)
	end
	
	local ratioA = sin((1 - delta) * halfTheta) / sinHalfTheta
	local ratioB = sin(delta * halfTheta) / sinHalfTheta
	
	local w = qa_w * ratioA + qb_w * ratioB
	local x = qa_x * ratioA + qb_x * ratioB
	local y = qa_y * ratioA + qb_y * ratioB
	local z = qa_z * ratioA + qb_z * ratioB
	
	return quaternion(w, x, y, z)
end

function math.clamp(x, min, max)
	return x >= max and max or (x <= min and min or x)
end

-- 
function math.smoothstep(a, b, x)
	x = math.clamp((x - a) / (b - a), 0, 1)
	return x * x * (3 - 2*x)
end

-- linear interpolation
function math.lerp(a, b, x)
	return a + (b - a) * x
end
